/*  
based on code from the TLC TLC5940 Library

Orginal Copyright (c) 2009 by Alex Leone <acleone ~AT~ gmail.com>

    This file is part of the Arduino TLC5940 Library.

    The Arduino TLC5940 Library is free software: you can redistribute it
    and/or modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    The Arduino TLC5940 Library is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with The Arduino TLC5940 Library.  If not, see
    <http://www.gnu.org/licenses/>. 
    
 */

//#ifndef LED_FADES_H
//#define LED_FADES_H

/** \file
    TLC fading functions. */

//#include <avr/interrupt.h>

//#include "Tlc5940.h"
//#include "WProgram.h"

#ifndef LED_FADE_BUFFER_LENGTH
/** The default fade buffer length (24).  Uses 24*13 = 312 bytes of ram. */
#define LED_FADE_BUFFER_LENGTH    64
#endif

/** Data for a single fade */
struct LED_Fade {
    uint8_t channel; /**< channel this fade is on */
    // 1 fade red
    // 2 fade green
    // 3 fade blue
    // 4 fade white
    // 5 fade hue
    // 6 fade sateration
    // 7 fade brightness
    uint8_t type; /**< channel this fade is on */
    int16_t startValue;       /**< value when the fade starts (0 - 4095) */
    int16_t changeValue;      /**< start + changeValue = endValue (0 - 4095) */
    uint32_t startMillis;     /**< millis() when to start */
    uint32_t endMillis;       /**< millis() when to end */
} LED_fadeBuffer[LED_FADE_BUFFER_LENGTH];

/** The current fade buffer size */
uint8_t LED_fadeBufferSize;

uint8_t LED_updateFades();
uint8_t LED_updateFades(uint32_t currentMillis);
//uint8_t LED_addFade(struct LED_Fade *fade);
//uint8_t LED_addFade(uint8_t channel, int16_t startValue, int16_t endValue, uint32_t startMillis, uint32_t endMillis);
uint8_t LED_isFading(uint8_t channel);
uint8_t LED_removeFades(uint8_t channel);
static void LED_removeFadeFromBuffer(LED_Fade *current, LED_Fade *end);

/** \addtogroup ExtendedFunctions
    \code #include "LED_fades.h" \endcode
     - uint8_t LED_updateFades() - updates all fades
     - uint8_t LED_updateFades(uint32_t currentMillis) - updates fades using
            currentMillis as the current time
     - uint8_t LED_addFade(struct LED_Fade *fade) - copies fade into the
            fade buffer
     - uint8_t LED_addFade(BYTEchannel, int16_t startValue,
            int16_t endValue, uint32_t startMillis, uint32_t endMillis) - adds
            a fade to the fade buffer
     - uint8_t LED_isFading(BYTEchannel) - returns 1 if there's
            a fade on this channel in the buffer
     - uint8_t LED_removeFades(BYTEchannel) - removes all fades
            on channel */
/* @{ */

/** Adds a fade to the buffer.
    \param fade the fade to be copied into the buffer
    \returns 0 if the fade buffer is full, fadeBufferSize if added successfully
*//*
uint8_t LED_addFade(struct LED_Fade *fade)
{
    if (LED_fadeBufferSize == LED_FADE_BUFFER_LENGTH) {
        return 0; // fade buffer full
    }
    struct LED_Fade *p = LED_fadeBuffer + LED_fadeBufferSize++;
    p->channel = fade->channel;
    p->type = fade->type;
    p->startValue = fade->startValue;
    p->changeValue = fade->changeValue;
    p->startMillis = fade->startMillis;
    p->endMillis = fade->endMillis;
    return LED_fadeBufferSize;
}
*/
/** Adds a fade to the fade buffer.
    \param channel the ouput channel this fade is on
    \param type of Fade
    \param startValue the value at the start of the fade
    \param endValue the value at the end of the fade
    \param startMillis the millis() when to start the fade
    \param endMillis the millis() when to end the fade
    \returns 0 if the fade buffer is full, fadeBufferSize if added successfully
*/

uint8_t LED_addFade(uint8_t channel, uint8_t type, int16_t startValue, int16_t endValue, uint32_t startMillis, uint32_t endMillis)
{
    if (LED_fadeBufferSize == LED_FADE_BUFFER_LENGTH) {
        return 0; // fade buffer full
    }
    struct LED_Fade *p = LED_fadeBuffer + LED_fadeBufferSize++;
    p->channel = channel;
    p->type = type;
    p->startValue = startValue;
    p->changeValue = endValue - startValue;
    p->startMillis = startMillis;
    p->endMillis = endMillis;
    return LED_fadeBufferSize;
}

/** Checks to see if any fades are happening on channel
    \param channel the channel to check
    \returns 1 if there is a fade in the buffer on this channel, 0 otherwise */
    
uint8_t LED_isFading(uint8_t channel)
{
    struct LED_Fade *end = LED_fadeBuffer + LED_fadeBufferSize;
    for (struct LED_Fade *p = LED_fadeBuffer; p < end; p++) {
        if (p->channel == channel) {
            return 1;
        }
    }
    return 0;
}

/** Removes any fades from the fade buffer on this channel.
    \param channel which channel the fades are on
    \returns how many fades were removed */
uint8_t LED_removeFades(uint8_t channel)
{
    uint8_t removed = 0;
    struct LED_Fade *end = LED_fadeBuffer + LED_fadeBufferSize;
    for (struct LED_Fade *p = LED_fadeBuffer; p < end; p++) {
        if (p->channel == channel) {
            removed++;
            LED_removeFadeFromBuffer(p, --end);
        }
    }
    return removed;
}

/** Copies the end of the buffer to the current and decrements
    LED_fadeBufferSize.  This will change the end of the buffer (pass by
    reference)
    \param current the fade to be removed
    \param endp the end of the fade buffer (pointer to pointer) */
static void LED_removeFadeFromBuffer(struct LED_Fade *current, struct LED_Fade *endp)
{
    if (endp != current) { // if this is not the last fade
        current->channel = endp->channel;
        current->startValue = endp->startValue;
        current->changeValue = endp->changeValue;
        current->startMillis = endp->startMillis;
        current->endMillis = endp->endMillis;
    }
    LED_fadeBufferSize--;
}

/** Updates fades using millis()
    \returns 0 if there are no fades left in the buffer. */
uint8_t LED_updateFades()
{
    return LED_updateFades(millis());
}

/** Updates any running fades.
    \param currentMillis the current millis() time.
    \returns 0 if there are no fades left in the buffer. */
uint8_t LED_updateFades(uint32_t currentMillis)
{
    struct LED_Fade *end = LED_fadeBuffer + LED_fadeBufferSize;
    //uint8_t needsUpdate = 0;
    for (struct LED_Fade *p = LED_fadeBuffer; p < end;){
        if (currentMillis >= p->endMillis) { // fade done
              HSB color;
             
              switch (p->type) {
              
              case 5:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {p->startValue + p->changeValue,color.s,color.b});
                break;
              case 6:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {color.h,p->startValue + p->changeValue,color.b});
                break;
             case 7:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {color.h,color.s,p->startValue + p->changeValue});
                break;
                
             case 1:
                leds[p->channel].r = p->startValue + p->changeValue;
                break;
              case 2:
                leds[p->channel].g = p->startValue + p->changeValue;
                break;
              case 3:  
                leds[p->channel].b = p->startValue + p->changeValue;
                break;
              case 4:  
                leds[p->channel].w = p->startValue + p->changeValue;
                break;    
             
            }  
            LED_removeFadeFromBuffer(p, --end);
            continue;
        } else {
            uint32_t startMillis = p->startMillis;
            if (currentMillis >= startMillis) {
              //update
                HSB color;
             
              switch (p->type) {
              
              case 5:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis),color.s,color.b});
                break;
              case 6:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {color.h,p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis),color.b});
                break;
             case 7:
                color = RGBToHSB(leds[p->channel]);
                leds[p->channel] = HSBToRGB((HSB) {color.h,color.s,p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis)});
                break;
                
             case 1:
                leds[p->channel].r = p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis);
                break;
              case 2:
                leds[p->channel].g = p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis);
                break;
              case 3:  
                leds[p->channel].b = p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis);
                break;
              case 4:  
                leds[p->channel].w = p->startValue + p->changeValue * (int32_t)(currentMillis - startMillis) / (int32_t)(p->endMillis - startMillis);
                break;    
             
            }  
        }
        p++;
    }}
    return LED_fadeBufferSize;
}

/* @} */
